# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.error import base


class StreamError(base.Error):
    MESSAGE = "Suanpan Stream Error: {}"


class StreamUnknownHandlerError(StreamError):
    MESSAGE = "Suanpan Stream Unknown Handler Error: {}"


class StreamUnknownHandlerClassError(StreamError):
    MESSAGE = "Suanpan Stream Unknown Handler Class Error: {}"


class StreamInvalidHandlerError(StreamError):
    MESSAGE = "Suanpan Stream Invalid Handler Error: {}"


class StreamHandlerExistsError(StreamError):
    MESSAGE = "Suanpan Stream Handler Exists Error: {}"


class StreamSIOError(StreamError):
    MESSAGE = "Suanpan Stream SIO Error: {}"



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import path
from suanpan.components import Result
from suanpan.storage.arguments import Folder


class CommonModel(Folder):
    def __init__(self, key, model=None, version="latest", **kwargs):
        super().__init__(key, **kwargs)
        self.version = version
        self.model = model

    def load(self, args):
        super().load(args)
        if self.folderPath:
            self.value = self.model
        return self

    def transform(self, value):
        if "component" in self.folderName:
            return self.model.loadFrom(self.folderName, version=self.version)

        folderPath = super().transform(value)
        if not folderPath:
            return None

        return self.model.load(folderPath)

    def clean(self):
        super().clean()
        if self.folderPath:
            self.value = self.model
        return self

    def save(self, result):
        model = result.value
        path.mkdirs(self.folderPath)
        model.save(self.folderPath)
        return super().save(Result.froms(value=self.folderPath))


class HotReloadModel(CommonModel):
    def transform(self, value):
        if not self.folderName:
            return None
        self.model.loadFrom(self.folderName, version=self.version)
        return self.model

    def save(self, result):
        raise NotImplementedError("Not support save!")



# coding=utf-8
from __future__ import absolute_import, print_function

import os

from suanpan import error
from suanpan import path as spath
from suanpan.app import app
from suanpan.app.modules.base import Module
from suanpan.storage import storage

module = Module()


@module.stream("storage.create")
def createStorage(context):
    args = context.args
    if not args.key:
        raise error.AppStorageModuleError("key is required")
    stream = context.stream
    stream.pipe(debugpath(args.key))
    if stream.done:
        upload(args.key)


@module.on("storage.download")
def downloadStrorage(context):
    args = context.args
    if not args.key:
        raise error.AppStorageModuleError("key is required")
    download(args.key)


@module.on("storage.upload")
def uploadStrorage(context):
    args = context.args
    if not args.key:
        raise error.AppStorageModuleError("key is required")
    upload(args.key)


@module.on("storage.remove")
def removeStrorage(context):
    args = context.args
    if not args.key:
        raise error.AppStorageModuleError("key is required")
    remove(args.key)


@module.on("storage.copy")
def copyStrorage(context):
    args = context.args
    if not args.src:
        raise error.AppStorageModuleError("src is required")
    if not args.dist:
        raise error.AppStorageModuleError("dist is required")
    copy(args.src, args.dist)


@module.on("storage.list")
def listStorage(context):
    args = context.args
    if not args.key:
        raise error.AppStorageModuleError("key is required")
    return [meta(key) for key in storage.listFiles(args.key)]


@module.on("storage.meta.get")
def getStorageMeta(context):
    args = context.args
    if not args.key:
        raise error.AppStorageModuleError("key is required")
    return meta(args.key)


@module.fn
def notifyMeta(key):
    app.sio.emit("storage.meta.notify", meta(key))


@module.fn
def meta(key):
    file = debugpath(key)
    mtime = os.path.getmtime(file) if os.path.exists(file) else None
    return {"src": f"debug/{key}", "key": key, "mtime": mtime}


@module.fn
def notify(data):
    app.sio.emit("storage.notify", data)


@module.fn
def download(key):
    notify({"key": key, "state": "downloading"})
    path = storage.download(
        key,
        debugpath(key),
        progress=lambda p: notify(
            {"key": key, "state": "downloading", "progress": p.progress}
        ),
    )
    notify({"key": key, "state": "downloaded"})
    notifyMeta(key)
    return path


@module.fn
def upload(key):
    notify({"key": key, "state": "uploading"})
    path = storage.upload(
        key,
        debugpath(key),
        progress=lambda p: notify(
            {"key": key, "state": "uploading", "progress": p.progress}
        ),
    )
    notify({"key": key, "state": "uploaded"})
    return path


@module.fn
def copy(src, dist):
    notify({"src": src, "dist": dist, "state": "copying"})
    storage.copy(
        src,
        dist,
        progress=lambda p: notify(
            {"src": src, "dist": dist, "state": "copying", "progress": p.progress}
        ),
    )
    notify({"src": src, "dist": dist, "state": "copied"})
    return dist


@module.fn
def remove(key):
    notify({"key": key, "state": "removing"})
    spath.remove(debugpath(key))
    storage.remove(
        key,
        progress=lambda p: notify(
            {"key": key, "state": "removing", "progress": p.progress}
        ),
    )
    notify({"key": key, "state": "removed"})
    return key


@module.fn
def debugpath(key):
    return app.sio.static("debug", key)



# coding=utf-8
from __future__ import absolute_import, print_function

import functools

from suanpan import g
from suanpan.proxy import Proxy


class Device(Proxy):
    MAPPING = {
        "zmq": "suanpan.device.zmq.ZMQDevice",
    }



# coding=utf-8
from __future__ import absolute_import, print_function

import abc
import os
import sys
import traceback
import uuid
import gc
import logging
import gevent
import gevent.pywsgi
import gevent.util
import socketio
from engineio import static_files
from gevent._greenlet_primitives import get_reachable_greenlets

from suanpan import asyncio, error, g, runtime, utils
from suanpan.arguments import Bool, BoolOrInt, Float, Int, String
from suanpan.components import Arguments
from suanpan.interfaces import HasDevMode, HasExitHooks, HasInitHooks, HasLogger, HasEventHooks
from suanpan.interfaces.optional import HasBaseServices
from suanpan.log import logger
from suanpan.mq import mq
from suanpan.mstorage import mstorage
from suanpan.storage import storage
from suanpan.stream import handlers, loops
from suanpan.stream.interfaces import HasHandlers
from suanpan.stream.objects import Context
from suanpan.stream.sio import StreamServer, WebSocketHandler
from suanpan.utils import functional, json, port
from suanpan.path import rootDir


class Stream(
    HasBaseServices, HasLogger, HasDevMode, HasInitHooks, HasExitHooks, HasHandlers, HasEventHooks
):
    STREAM_ARGUMENTS = [
        String("stream-recv-queue", default=f"mq-{g.nodeId}"),
        BoolOrInt("stream-recv-queue-block", default=60000),
        Float("stream-recv-queue-delay", default=0),
        Int("stream-recv-queue-max-length", default=1000),
        Bool("stream-recv-queue-trim-immediately", default=False),
        Bool("stream-recv-queue-retry", default=False),
        Int("stream-recv-queue-retry-max-count", default=100),
        Float("stream-recv-queue-retry-timeout", default=1.0),
        Int("stream-recv-queue-retry-max-times", default=3),
        String("stream-send-queue", default="mq-master"),
        Int("stream-send-queue-max-length", default=1000),
        Bool("stream-send-queue-trim-immediately", default=False),
        String("stream-sio-static-folder", default="/web"),
        String("stream-sio-static-url-prefix", default="/"),
        String("stream-app-event-hub", default=f"app-event-hub-{g.appId}"),
        Int("stream-app-event-hub-max-length", default=1000),
    ]
    STREAM_PARAMS_FILE = "params.json"

    def __init__(self):
        super().__init__()
        self.title = ""
        self.argsDict = {}
        self.options = None
        self.args = None
        self._exit_code = 0

        self.callLoop = StreamCallLoop(self)
        self.triggerLoop = StreamTriggerLoop(self)
        self.sioLoop = StreamSIOLoop(self)
        self.httpLoop = StreamHTTPLoop(self)
        self.pubsubLoop = StreamPubsubLoop(self)

    @property
    def loops(self):
        return [self.callLoop, self.triggerLoop, self.sioLoop, self.httpLoop, self.pubsubLoop]

    def beforeInit(self):
        logger.logDebugInfo()
        logger.debug(f"Stream {self.name} starting...")

    def init(self, *args, **kwargs):
        self.argsDict = self.getArgsDict(*args, **kwargs)
        self.options = self.loadGlobalArguments(self.argsDict)
        self.setBaseServices(self.options)
        self.args = self.loadComponentArguments(self.argsDict)
        return Context.froms(args=self.args)

    def afterInit(self, _):
        self.updateParams(self.loadParams())

    def loadGlobalArguments(self, argsDict=None):
        logger.debug("Loading Global Arguments:")
        argsDict = argsDict or self.argsDict
        args = self.loadFormatArguments(self.getGlobalArguments(), argsDict)
        return Arguments.froms(*args)

    def loadComponentArguments(self, argsDict=None):
        logger.debug("Loading Components Arguments:")
        argsDict = argsDict or self.argsDict
        args = self.loadFormatArguments(self.getComponentArguments(), argsDict)
        return Arguments.froms(*args)

    def getGlobalArguments(self, *args, **kwargs):
        arguments = super().getGlobalArguments(*args, **kwargs)
        return arguments + self.STREAM_ARGUMENTS

    def getParams(self):
        return {arg.paramName: self.args.get(arg.paramName) for arg in self.ARGUMENTS}

    def updateParams(self, params):
        for key, value in params.items():
            arg = self.paramsArgsMap.get(key)
            if arg and value is not None:
                self.args.update({key: value for key in (arg.key, arg.alias) if key})
        return self.getParams()

    def loadParams(self):
        logger.debug(f"Loading params from {self.paramsFilePath}")
        try:
            storage.download(self.paramsFileKey, self.paramsFilePath)
            params = json.load(self.paramsFilePath)
            logger.debug(f"Loaded params: {params}")
            return params
        except Exception:  # pylint: disable=broad-except
            # tracebackInfo = traceback.format_exc()
            # logger.debug(f"Load params error: {tracebackInfo}")
            return {}

    def saveParams(self, params):
        logger.debug(f"Saving params: {self.paramsFilePath}")
        json.dump(params, self.paramsFilePath)
        storage.upload(self.paramsFileKey, self.paramsFilePath)
        return params

    @functional.lazyproperty
    def paramsArgsMap(self):
        argsmap = {}
        for arg in self.ARGUMENTS:
            if arg.key:
                argsmap[arg.key] = arg
            if arg.alias:
                argsmap[arg.alias] = arg
        return argsmap

    @functional.lazyproperty
    def paramsFileKey(self):
        return storage.getKeyInNodeConfigsStore(self.STREAM_PARAMS_FILE)

    @functional.lazyproperty
    def paramsFilePath(self):
        return storage.getPathInNodeConfigsStore(self.STREAM_PARAMS_FILE)

    @runtime.globalrun
    def start(self, *args, **kwargs):
        self.callBeforeInitHooks()
        context = self.init(*args, **kwargs)
        self.callAfterInitHooks(context)
        self.registerBeforeExitHooks(context)
        self.run()

    def run(self):
        runners = (loop.start for loop in self.loops if loop.ready)
        asyncio.wait(asyncio.run(runners))
        logger.debug("Stream ALL Loop Closed")

        if self._exit_code != 0:
            logging.shutdown()
            # gevent.util.print_run_info()

            # glets = get_reachable_greenlets()
            # running = []
            # for ob in glets:
            #     if isinstance(ob, gevent.Greenlet) and bool(ob):
            #         running.append(ob)

            # kill greenlet created by other modules
            # running = [obj for obj in gc.get_objects() if isinstance(obj, gevent.Greenlet) and bool(obj)]
            # gevent.killall(running)
            sys.exit(self._exit_code)

    def close(self, code=0):
        self._exit_code = code
        for loop in self.loops:
            if loop.ready:
                loop.close()

    @property
    def vars(self):
        return mstorage.vars


class StreamBaseLoop(HasHandlers):
    __metaclass__ = abc.ABCMeta

    DEFAULT_HANDLER_KEY = None
    DEFAULT_HANDLER_CLASS = handlers.Handler
    DEFAULT_LOGGER_MAX_LENGTH = 120
    DEFAULT_MESSAGE = {}

    def __init__(self, stream):
        super().__init__()
        self.stream = stream

    @property
    def ready(self):
        return self.hasHandler(self.DEFAULT_HANDLER_KEY)

    def getHandler(self, key):
        if not self.hasHandler(key):
            super().setHandler(key, self.DEFAULT_HANDLER_CLASS())
        return super().getHandler(key)

    def generateRequestId(self):
        return uuid.uuid4().hex

    def generateMessage(self, **kwargs):
        message = {}
        message.update(self.DEFAULT_MESSAGE, **kwargs)
        message["id"] = self.generateRequestId()
        return message

    @abc.abstractmethod
    def start(self):
        pass


class StreamMQBaseLoop(StreamBaseLoop):
    def _send(self, message, data, queue=None):
        queue = queue or self.stream.options["stream-send-queue"]
        data = {
            "node_id": g.nodeId,
            "request_id": message["id"],
            **data,
        }
        extra = message.get("extra")
        if extra:
            data.update(extra=extra)
        logger.debug(
            utils.shorten(f"Send to `{queue}`: {data}", self.DEFAULT_LOGGER_MAX_LENGTH)
        )
        return mq.sendMessage(
            queue,
            data,
            maxlen=self.stream.options["stream-send-queue-max-length"],
            trimImmediately=self.stream.options["stream-send-queue-trim-immediately"],
        )

    def sendSuccessMessage(self, message, data, queue=None):
        if not all(key.startswith("out") for key in data):
            raise error.StreamError(
                "Success Message data only accept keys starts with 'out'"
            )
        data = {key: value for key, value in data.items() if value is not None}
        data.update(success="true")
        return self._send(message, data, queue=queue)

    def sendFailureMessage(self, message, msg, queue=None):
        if not isinstance(msg, str):
            raise error.StreamError("Failure Message msg only accept string")
        data = {"msg": msg, "success": "false"}
        return self._send(message, data, queue=queue)

    def send(self, results, queue=None, message=None, args=None, **_):
        message = message or self.generateMessage()
        outputs = self.getHandler().save(results, args=args, message=message)
        if outputs:
            self.sendSuccessMessage(message, outputs, queue=queue)

    def sendError(self, msg, queue=None, message=None):
        message = message or self.generateMessage()
        return self.sendFailureMessage(message, msg, queue=queue)

    def sendMissionMessage(self, message, data, queue=None):
        if not all(key.startswith("in") for key in data):
            raise error.StreamError(
                "Mission Message data only accept keys starts with 'in'"
            )
        data = {key: value for key, value in data.items() if value is not None}
        data.update(id=message["id"])
        return self._send(message, data, queue=queue)

    def getHandler(self):
        if not self.hasHandler(self.DEFAULT_HANDLER_KEY):
            self.setHandler(self.DEFAULT_HANDLER_CLASS())
        return super().getHandler(self.DEFAULT_HANDLER_KEY)

    def setHandler(self, handler):
        return super().setHandler(
            self.DEFAULT_HANDLER_KEY, handler
        )

    def handle(self, message, *args, **kwargs):
        title = f"{message['id']} - {self.DEFAULT_HANDLER_KEY}"
        _run = runtime.costrun(title)(self.getHandler().run)

        try:
            outputs = _run(self.stream, message, *args, **kwargs)
            if outputs:
                self.sendSuccessMessage(message, outputs)
        except Exception as e:  # pylint: disable=broad-except
            tracebackInfo = traceback.format_exc()
            logger.error(tracebackInfo, extra={"success": "false"})
            self.sendFailureMessage(message, tracebackInfo)
            self.stream.sioLoop.notifyError(e)


class StreamCallLoop(StreamMQBaseLoop):
    DEFAULT_HANDLER_KEY = "call"
    DEFAULT_HANDLER_CLASS = handlers.CallHandler

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.lastMessage = None

    def subscribe(self, **kwargs):
        return mq.subscribeQueue(
            self.stream.options["stream-recv-queue"],
            group=g.nodeGroup,
            consumer=g.nodeId,
            block=self.stream.options["stream-recv-queue-block"],
            delay=self.stream.options["stream-recv-queue-delay"],
            **kwargs,
        )

    def retryPendingMessages(self, **kwargs):
        return mq.retryPendingMessages(
            self.stream.options["stream-recv-queue"],
            group=g.nodeGroup,
            consumer=g.nodeId,
            count=self.stream.options["stream-recv-queue-retry-max-count"],
            maxTimes=self.stream.options["stream-recv-queue-retry-max-times"],
            timeout=self.stream.options["stream-recv-queue-retry-timeout"],
            maxlen=self.stream.options["stream-recv-queue-max-length"],
            trimImmediately=self.stream.options["stream-recv-queue-trim-immediately"],
            **kwargs,
        )

    def retryLastMessage(self):
        if not self.lastMessage:
            logger.debug("Stream Call Loop No Message To Retry")
            return

        self._send(
            self.lastMessage,
            self.lastMessage,
            queue=self.stream.options["stream-recv-queue"],
        )

    def close(self):
        logger.debug("Stream Call Loop Closing...")
        mq.close()
        logger.debug("Stream Call Loop Closed")

    @runtime.globalrun
    def start(self):
        logger.debug("Stream Call Loop Starting...")
        if self.stream.options["stream-recv-queue-retry"]:
            self.retryPendingMessages()
        for message in self.subscribe():
            self.lastMessage = message.get("data", {})
            self.handle(self.lastMessage)


class StreamTriggerLoop(StreamMQBaseLoop):
    DEFAULT_HANDLER_KEY = "trigger"
    DEFAULT_HANDLER_CLASS = handlers.TriggerHandler
    DEFAULT_LOOP = loops.Interval(1)

    def __init__(self, stream, loop=None):
        super().__init__(stream)
        self._loop = loop or self.DEFAULT_LOOP

    def _list(self, data):
        if isinstance(data, (tuple, list)):
            return data
        if data is None:
            return []
        return [data]

    def loop(self, funcOrIter, *_, **__):
        self._loop = funcOrIter if callable(funcOrIter) else lambda: funcOrIter

    def close(self):
        logger.debug("Stream Trigger Loop Closing...")
        self._loop.close()
        logger.debug("Stream Trigger Loop Closed")

    @runtime.globalrun
    def start(self):
        logger.debug("Stream Trigger Loop Starting...")
        for data in self._loop():
            self.handle(self.generateMessage(), *self._list(data))


class StreamSIOLoop(StreamBaseLoop):
    DEFAULT_HANDLER_CLASS = handlers.SIOHandler
    EVENT_CONNECT = "connect"
    EXTRA_CONTEXT_TYPES = {"mjs": "text/javascript", "svg": "image/svg+xml"}

    def __init__(self, stream):
        super().__init__(stream)
        self.sio = socketio.Server(async_mode="gevent", cors_allowed_origins="*", json=json,
                                   max_http_buffer_size=100000000)
        self.options = self._getDefaultOptions()
        self.extraStaticFiles = {}
        self.staticFiles = {}
        self._enabled = not g.serviceSioDisable
        self.server = None
        self.web = None

    def _getDefaultOptions(self):
        options = Context()
        return options

    @functional.lazyproperty
    def staticFolder(self):
        root = rootDir()
        staticF = self.stream.options["stream-sio-static-folder"]
        if root:
            return os.path.join(root, os.path.relpath(staticF, "/"))
        return staticF

    @functional.lazyproperty
    def staticUrlPrefix(self):
        return self.getFullUrlPrefix(
            self.stream.options["stream-sio-static-url-prefix"]
        )

    @functional.lazyproperty
    def siostream(self):
        return StreamServer(self.stream, self.sio)

    @property
    def contentTypes(self):
        return static_files.content_types

    @property
    def ready(self):
        return self._enabled or g.debug

    def enable(self):
        self._enabled = True
        return self

    def disable(self):
        self._enabled = False
        return self

    def getHandler(self, key):
        initial = not self.hasHandler(key)
        handler = super().getHandler(key)
        if initial:
            self.sio.on(key)(self._handle(key, handler))
        return handler

    def setHandler(self, key, handler):
        return self.getHandler(key).use(handler)

    def _handle(self, key, handler):
        def _dec(sid, data):
            title = f"{sid} - sio:{key}"
            _run = runtime.costrun(title)(handler.run)
            try:
                result = _run(self.stream, data, sid)
                return {"success": True, "data": result}
            except Exception as e:  # pylint: disable=broad-except
                logger.error(traceback.format_exc())
                return {"success": False, "error": self.formatError(e)}

        return _dec

    def getDefaultStaticFiles(self, folder):
        self.contentTypes.update(self.EXTRA_CONTEXT_TYPES)
        folder = self.getFullUrlPrefix(os.path.abspath(folder))
        prefixUrl = self.getFullUrlPrefix(self.staticUrlPrefix)
        return {prefixUrl: folder, **self.extraStaticFiles}

    def getFullUrlPrefix(self, prefix):
        return prefix.rstrip("/") + "/"

    def emit(self, *args, **kwargs):
        if self.ready:
            self.sio.emit(*args, **kwargs)

    def enter(self, *args, **kwargs):
        self.sio.enter_room(*args, **kwargs)

    def leave(self, *args, **kwargs):
        self.sio.leave_room(*args, **kwargs)

    def notifyError(self, err, *args, **kwargs):
        self.emit("errors.notify", {"error": self.formatError(err)}, *args, **kwargs)

    def formatError(self, err):
        return {
            "message": str(err),
            "traceback": traceback.format_exc(),
        }

    def close(self):
        logger.debug("Stream SIO Loop Closing...")
        if self.server:
            self.server.close()
        logger.debug("Stream SIO Loop Closed")

    def setWebApp(self, web):
        self.web = web

    def _init_win_server(self):
        p = None
        for _ in range(20):
            try:
                p = port.get_free_port()
                self.server.set_listener(("0.0.0.0", p))
                self.server.start()
                logger.info("Stream SIO Loop listen port {}".format(p))
                port.register_server(g.port, p)
                return
            except OSError as e:
                if e.errno in [10013, 10048]:
                    logger.warn("Stream SIO Loop listen port conflict {}".format(p))
                    continue

                raise
            except Exception:
                raise

        raise Exception("port conflicts repeatedly failed, please try again later")

    def _start_server(self):
        if port.need_free_port():
            self._init_win_server()
        else:
            self.server.start()

    @runtime.globalrun
    def start(self):
        logger.debug("Stream SIO Loop Starting...")
        self.staticFiles = self.getDefaultStaticFiles(self.staticFolder)
        if self.web:
            app = self.web
        else:
            app = socketio.WSGIApp(self.sio, static_files=self.staticFiles)
        self.server = gevent.pywsgi.WSGIServer(
            ("", g.port), app, handler_class=WebSocketHandler
        )
        self._start_server()
        self.server.serve_forever()


class StreamHTTPLoop(StreamBaseLoop):
    def __init__(self, stream):
        super().__init__(stream)
        self._enabled = not g.serviceHttpDisable
        self.server = None

    @property
    def ready(self):
        return self._enabled or g.debug

    def enable(self):
        self._enabled = True
        return self

    def disable(self):
        self._enabled = False
        return self

    def application(self, env, start_response):
        if env['REQUEST_METHOD'] == 'POST' and env['PATH_INFO'] == '/internal/trap':
            start_response('200 OK', [('Content-Type', 'text/json')])
            self.stream.close(code=98)
            return [json.dumps({"success": "true"}, ensure_ascii=False).encode('utf8')]

        start_response('404 Not Found', [('Content-Type', 'text/json')])
        return [b'{"success": "false", "msg": "invalid request"}']

    def close(self):
        logger.debug("Stream HTTP Loop Closing...")
        if self.server:
            self.server.close()
        logger.debug("Stream HTTP Loop Closed")

    def _init_win_server(self):
        p = None
        for _ in range(20):
            try:
                p = port.get_free_port()
                self.server.set_listener(("0.0.0.0", p))
                self.server.start()
                logger.info("Stream HTTP Loop listen port {}".format(p))
                port.register_server(g.termPort, p)
                return
            except OSError as e:
                if e.errno in [10013, 10048]:
                    logger.warn("Stream HTTP Loop listen port conflict {}".format(p))
                    continue

                raise
            except Exception:
                raise

        raise Exception("port conflicts repeatedly failed, please try again later")

    def _start_server(self):
        if port.need_free_port():
            self._init_win_server()
        else:
            self.server.start()

    @runtime.globalrun
    def start(self):
        logger.debug("Stream HTTP Loop Starting...")
        self.server = gevent.pywsgi.WSGIServer(("", g.termPort), self.application)
        self._start_server()
        self.server.serve_forever()


class StreamPubsubLoop(StreamBaseLoop):
    def __init__(self, stream):
        super().__init__(stream)

    @property
    def ready(self):
        return True

    def subscribe(self, **kwargs):
        return mq.subscribeEvent(
            self.stream.options["stream-app-event-hub"],
            block=self.stream.options["stream-recv-queue-block"],
            delay=self.stream.options["stream-recv-queue-delay"],
            **kwargs,
        )

    def close(self):
        logger.debug("Stream Pubsub Loop Closed")

    @runtime.globalrun
    def start(self):
        logger.debug("Stream Pubsub Loop Starting...")
        for event in self.subscribe():
            _event = event.get("data", {})
            self.handle(_event)

    def handle(self, event):
        topic = event.get('event_name')
        if not topic:
            return

        data = event.get('data')
        if not data:
            return

        try:
            handler = self.stream.eventHooks.get(topic)
            if handler:
                logger.debug(f"handle event: {event}")
                d = json.loads(data)
                handler(d)
        except Exception:
            logger.error(f"handle event {topic} error:", exc_info=True)

    @runtime.retry(stop_max_attempt_number=3)
    def send(self, topic: str, payload: dict):
        data = {"app_id": g.appId, "node_id": g.nodeId, "event_name": topic, "data": payload}
        logger.debug(
            utils.shorten(f"Send event: {data}", self.DEFAULT_LOGGER_MAX_LENGTH)
        )
        return mq.sendEvent(
            self.stream.options["stream-app-event-hub"],
            data,
            maxlen=self.stream.options["stream-app-event-hub-max-length"],
            trimImmediately=self.stream.options["stream-send-queue-trim-immediately"],
        )



# coding=utf-8
from __future__ import absolute_import, print_function

from contextlib import contextmanager

from mysql import connector

from suanpan.dw import base


class DataWarehouse(base.RelationalDataWarehouse):
    def __init__(  # pylint: disable=useless-super-delegation
        self,
        mysqlHost,
        mysqlPort,
        mysqlUsername,
        mysqlPassword,
        mysqlDatabase,
        **kwargs
    ):
        super().__init__(
            mysqlHost, mysqlPort, mysqlUsername, mysqlPassword, mysqlDatabase, **kwargs
        )

    @contextmanager
    def connect(self):
        connection = connector.connect(
            host=self.host,
            port=self.port,
            database=self.database,
            user=self.user,
            password=self.password,
        )

        yield connection
        connection.close()

    def getColumnsNamesFromCursor(self, cursor):
        return cursor.column_names



# coding=utf-8
from __future__ import absolute_import, print_function

import itertools
import tempfile
import typing
from suanpan.arguments import Bool, Int, String
from suanpan.proxy import Proxy
from suanpan import error

# from suanpan.storage import local, minio, oss

DEFAULT_TEMP_STORE = tempfile.gettempdir()

if typing.TYPE_CHECKING:
    from suanpan.stypes import StorageT


class StorageProxy(Proxy):
    # MAPPING = {"oss": oss.Storage, "local": local.Storage, "minio": minio.Storage}
    DEFAULT_ARGUMENTS = [String("storage-type", default="oss")]
    OSS_ARGUMENTS = [
        String("storage-oss-access-id"),
        String("storage-oss-access-key"),
        String("storage-oss-bucket-name", default="suanpan"),
        String("storage-oss-endpoint", default="http://oss-cn-beijing.aliyuncs.com"),
        String("storage-oss-delimiter", default="/"),
        String("storage-oss-temp-store", default=DEFAULT_TEMP_STORE),
        Int("storage-oss-download-num-threads", default=1),
        String("storage-oss-download-store-name", default=".py-oss-download"),
        Int("storage-oss-upload-num-threads", default=1),
        String("storage-oss-upload-store-name", default=".py-oss-upload"),
    ]
    LOCAL_ARGUMENTS = [String("storage-local-temp-store", default=DEFAULT_TEMP_STORE)]
    MINIO_ARGUMENTS = [
        String("storage-minio-access-key"),
        String("storage-minio-secret-key"),
        String("storage-minio-bucket-name", default="suanpan"),
        String("storage-minio-endpoint"),
        Bool("storage-minio-secure", default=True),
        String("storage-minio-delimiter", default="/"),
        String("storage-minio-temp-store", default=DEFAULT_TEMP_STORE),
    ]
    ARGUMENTS = list(
        itertools.chain(
            DEFAULT_ARGUMENTS, OSS_ARGUMENTS, LOCAL_ARGUMENTS, MINIO_ARGUMENTS
        )
    )

    def importBackend(self, backendType):
        if backendType == "oss":
            from suanpan.storage import oss
            return oss.Storage
        elif backendType == "minio":
            from suanpan.storage import minio
            return minio.Storage
        elif backendType == "local":
            from suanpan.storage import local
            return local.Storage
        else:
            raise error.ProxyError(
                f"{self.name} don't supported backend type {backendType}"
            )


storage: 'StorageT' = StorageProxy()



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.error import base


class MStorageError(base.Error):
    MESSAGE = "Suanpan MStorage Error: {}"


class MStorageVarError(base.Error):
    MESSAGE = "Suanpan MStorage Var Error: {}"


class MStorageUnknownVarTypeError(base.Error):
    MESSAGE = "Suanpan MStorage Unknown Var Type: {}"


class MStorageUnregisterVarKeyError(base.Error):
    MESSAGE = "Suanpan MStorage Unregister Var Key: {}"


class MStorageVarHasRegisteredAsAnotherTypeError(base.Error):
    MESSAGE = "Suanpan MStorage Var Has Registered As Another Type: {}"



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import error, objects
from suanpan.stream import handlers


class HasHandlers(objects.HasName):
    def __init__(self):
        super().__init__()
        self.handlers = {}

    def addHandler(self, key, handler):
        if not isinstance(handler, handlers.Handler):
            raise error.StreamInvalidHandlerError(f"{self.name}.{key}")
        if self.hasHandler(key):
            raise error.StreamHandlerExistsError(f"{self.name}.{key}")
        self.setHandler(key, handler)
        return self

    def hasHandler(self, key):
        return key in self.handlers

    def getHandler(self, key):
        handler = self.handlers.get(key)
        if not handler:
            raise error.StreamUnknownHandlerError(f"{self.name}.{key}")
        return handler

    def setHandler(self, key, handler):
        self.handlers[key] = handler
        return self

    def removeHandler(self, key):
        handler = self.handlers.pop(key, None)
        if not handler:
            raise error.StreamUnknownHandlerError(f"{self.name}.{key}")
        return handler



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import asyncio, path
from suanpan.utils import convert


def read(file, *args, **kwargs):
    import cv2  # pylint: disable=import-outside-toplevel

    img = cv2.imread(file, *args, **kwargs)  # pylint: disable-msg=e1101
    if img is None:
        raise Exception(f"Image read failed: {file}")
    return img


def _save(file, image):
    import cv2  # pylint: disable=import-outside-toplevel

    path.safeMkdirsForFile(file)
    success = cv2.imwrite(file, image)  # pylint: disable-msg=e1101
    if not success:
        path.remove(file)
        raise Exception(f"Image save failed: {file}")
    return file


def _saveflat(file, data):
    image = convert.flatAsImage(data)
    return _save(file, image)


def _savegif(file, data):
    import imageio  # pylint: disable=import-outside-toplevel

    image3D = convert.to3D(data)
    path.safeMkdirsForFile(file)
    imageio.mimsave(file, image3D)
    return file


def save(file, data, flag=None):
    mapping = {None: _save, "flat": _saveflat, "animated": _savegif}
    func = mapping.get(flag)
    if not func:
        raise Exception(f"Unknow flag: {flag}")
    return func(file, data)


def saves(filepathPrefix, images):
    counts = len(images)
    n = len(str(counts))
    iterable = (
        (f"{filepathPrefix}_{str(index).zfill(n)}.png", image)
        for index, image in enumerate(images)
    )
    asyncio.starmap(_save, iterable)
    return filepathPrefix


def saveall(filepathPrefix, data):
    image3D = convert.to3D(data)
    layers = len(image3D)
    n = len(str(layers))
    runners = [
        asyncio.run(_save, f"{filepathPrefix}_{str(index).zfill(n)}.png", image)
        for index, image in enumerate(image3D)
    ]
    runners.append(asyncio.run(_saveflat, f"{filepathPrefix}.png", image3D))
    runners.append(asyncio.run(_savegif, f"{filepathPrefix}.gif", image3D))
    asyncio.wait(runners)
    return filepathPrefix


def resize(data, size):
    import cv2  # pylint: disable=import-outside-toplevel

    return cv2.resize(data, size)  # pylint: disable-msg=e1101



# coding=utf-8
from __future__ import absolute_import, print_function

import argparse
import contextlib
import copy
import faulthandler
import os
import sys

from suanpan.imports import imports
from suanpan.utils import env as spenv
from suanpan.helper import init_project, sio_send


def run(component, *args, **kwargs):
    if isinstance(component, str):
        if os.sep in component:
            component = component.lstrip(f'.{os.sep}')

        component = f"{component[:-3]}.app" if component.endswith(".py") else component
        component = imports(component)
    with env(**kwargs.pop("env", {})):
        return component.start(*args, **kwargs)


@contextlib.contextmanager
def env(**kwargs):
    old = copy.deepcopy(spenv.environ)
    spenv.update(kwargs)
    yield spenv.environ
    spenv.update(old)


def cli():
    sys.path.append(os.path.abspath(os.curdir))
    parser = argparse.ArgumentParser()
    parser.add_argument("component")
    _args, _rest = parser.parse_known_args()

    sys.argv = sys.argv[:1]
    return run(_args.component, *_rest)


def helper():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest='command', required=True, title='subcommands',
                                       help='suanpan helper command')
    parser_init = subparsers.add_parser('init')
    parser_init.add_argument('project', nargs='?', type=str, help='new component name')
    parser_init.add_argument('-f', '--format', choices=['normal', 'function', 'autoload'], default='normal',
                             help='init component with code format')
    parser_run = subparsers.add_parser('run')
    parser_run.add_argument('component', help='component name')
    parser_sio = subparsers.add_parser('sio')
    parser_sio.add_argument('event', help='sio event')
    parser_sio.add_argument('-s', '--data-string', default='{}', help='sio data string')
    parser_sio.add_argument('-f', '--data-file', help='sio data file')

    args, _rest = parser.parse_known_args()
    if args.command == 'init':
        init_project(args.project, args.format)
    elif args.command == 'sio':
        sio_send(args.event, args.data_string, args.data_file)
    else:
        sys.path.append(os.path.abspath(os.curdir))
        return run(args.component, *_rest)


if __name__ == "__main__":
    faulthandler.enable()
    cli()



# coding=utf-8
from __future__ import absolute_import, print_function

import suanpan
from suanpan.app import app
from suanpan.app.arguments import String
from suanpan.storage import storage


@app.param(String(key="local", required=True))
@app.param(String(key="remote", required=True))
def SPStorageUpload(context):
    args = context.args
    storage.upload(args.remote, args.local)


if __name__ == "__main__":
    suanpan.run(app)



# coding=utf-8
from __future__ import absolute_import, print_function

import typing
from suanpan import app
from suanpan.proxy import Proxy

if typing.TYPE_CHECKING:
    from suanpan.stypes import NodeT


class Node(Proxy):
    MAPPING = {
        "spark": "suanpan.node.spark.SparkNode",
        "docker": "suanpan.node.docker.DockerNode",
        "stream": "suanpan.node.stream.StreamNode",
    }


node: 'NodeT' = Node(type=app.TYPE)



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import error, path, runtime
from suanpan.state.base import IndexSaver, Saver, TimeSaver
from suanpan.storage import storage


class StorageSaver(Saver):
    PATTERN = None

    def __init__(self, name, **kwargs):
        super().__init__(storageName=name, **kwargs)

    @property
    def currentPattern(self):
        _pattern = self.current.pattern or self.PATTERN
        if not _pattern:
            raise error.StateError("Pattern is not set!")
        return _pattern.format(**self.current)

    def update(self):
        self.current.localPath = storage.getPathInTempStore(self.current.storageName)
        storageName = storage.storagePathJoin(
            self.current.storageName, self.currentPattern
        )
        localPath = storage.localPathJoin(self.current.localPath, self.currentPattern)
        return storage.upload(storageName, localPath)

    def clean(self):
        self.current.localPath = storage.getPathInTempStore(self.current.storageName)
        runtime.saferun(storage.remove)(self.current.storageName)
        return path.remove(self.current.localPath)


class StorageIndexSaver(StorageSaver, IndexSaver):
    pass


class StorageTimeSaver(StorageSaver, TimeSaver):
    pass



import abc


class ArgumentT(object):
    @abc.abstractmethod
    def __init__(self, key, alias=None, label=None, required=False, default=None, choices=[], type=str):
        """init an argument

        Arguments:
            key: the key of the argument
            alias: the alias name of the argument
            label: the label of the argument
            required: is this argument required
            default: default value of the argument
            choices: choices of the argument if have
            type: argument type

        Example:
            String(key="formula", default="")
        """
        ...



# coding=utf-8
from __future__ import absolute_import, print_function

import datetime
import io
import traceback
import urllib
import uuid

import geventwebsocket.handler

from suanpan import path, runtime
from suanpan.log import logger
from suanpan.stream import handlers
from suanpan.stream.interfaces import HasHandlers


class Event(object):
    TEMPLATE = "__{keyword}_::{event}::{data}__"

    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        self._kwargs.setdefault("keyword", "akuma")

    def __call__(self, event, *args, **kwargs):
        kwargs.update(event=event)
        return self.format(*args, **kwargs)

    def format(self, *args, **kwargs):
        return self.TEMPLATE.format(*self._args, *args, **self._kwargs, **kwargs)

    def new(self):
        return self.format(event="new", data="id")

    def resume(self, id=""):
        return self.format(event="resume", data=id)

    def data(self, id):
        return self.format(event="data", data=id)

    def more(self, id):
        return self.format(event="more", data=id)

    def stop(self):
        return self.format(event="stop", data="")

    def end(self, id):
        return self.format(event="end", data=id)


class Stream(object):
    def __init__(self, id):
        self.record = {
            "id": id,
            "event": None,
            "size": 0,
            "total": 0,
            "expire": self._addExipre(new=True),
            "data": {},
        }
        self.buffer = io.BytesIO()

    def _addExipre(self, expire=None, new=False):
        expire = expire or self.now
        if new:
            return expire + datetime.timedelta(minutes=5)
        if (self.now - expire).seconds <= 60:
            return expire + datetime.timedelta(milliseconds=60)
        return expire

    @property
    def now(self):
        return datetime.datetime.now()

    @property
    def sid(self):
        return self.record["sid"]

    @property
    def event(self):
        return self.record["event"]

    @property
    def size(self):
        return self.record["size"]

    @property
    def total(self):
        return self.record["total"]

    @property
    def dirty(self):
        return self.record["dirty"]

    @property
    def expire(self):
        return self.record["expire"]

    @property
    def data(self):
        return self.record["data"]

    @property
    def expired(self):
        return self.now > self.expire

    @property
    def done(self):
        return self.total >= self.size

    def update(self, *args, **kwargs):
        self.record.update(*args, **kwargs)
        return self

    def read(self):
        self.buffer.seek(0)
        data = self.buffer.read()
        self.buffer.truncate(0)
        self.buffer.seek(0)
        return data

    def write(self, chunks):
        if not chunks:
            return 0

        total = self.buffer.write(chunks)
        self.update(total=self.total + total, expire=self._addExipre(self.expire))
        return total

    def pipe(self, file):
        path.mkdirs(file, parent=True)
        chunks = self.read()
        if self.total - len(chunks) <= 0:
            path.remove(file)
        with open(file, "ab") as f:
            return f.write(chunks)


class Cleaner(object):
    def __init__(self, delta, handler):
        self.delta = delta
        self.handler = handler
        self.expire = self.now + delta

    @property
    def now(self):
        return datetime.datetime.now()

    @property
    def expired(self):
        return self.now > self.expire

    def clean(self, *args, **kwargs):
        if self.expired:
            self.handler(*args, **kwargs)


class StreamServer(HasHandlers):
    DEFAULT_HANDLER_CLASS = handlers.SIOStreamHandler

    def __init__(self, stream, sio):
        super().__init__()
        self.stream = stream
        self.sio = sio
        self.streams = {}
        self.event = Event()
        self.cleaner = Cleaner(datetime.timedelta(seconds=10), self._clean)
        self.init()

    def init(self):
        self.sio.on(self.event.new(), self.onNew)
        self.sio.on(self.event.resume(), self.onResume)

    @property
    def now(self):
        return datetime.datetime.now()

    def _addTime(self, date, new=False):
        if new:
            return date + datetime.timedelta(minutes=5)
        if (self.now - date).seconds <= 60:
            return date + datetime.timedelta(milliseconds=60)
        return date

    def _genid(self):
        id = uuid.uuid4().hex
        while id in self.streams:
            id = uuid.uuid4().hex
        return id

    def _createNew(self, id):
        self._listener(id)
        return id

    def _listener(self, id):
        self.sio.on(self.event.data(id), lambda sid, data: self.onData(sid, id, data))

    def _done(self, id, *args, **kwargs):
        self.end(id, *args, **kwargs)
        self.streams.pop(id)

    def _clean(self):
        expireids = [id for id, stream in self.streams.items() if stream.expired]
        for id in expireids:
            self.streams.pop(id)

    def _handle(self, key, handler):
        def _dec(sid, data, stream):
            title = f"{sid} - sio:{key}"
            _run = runtime.costrun(title)(handler.run)
            try:
                return _run(self.stream, data, sid, stream)
            except Exception:  # pylint: disable=broad-except
                logger.error(traceback.format_exc())
                return None

        return _dec

    def getHandler(self, key):
        if not self.hasHandler(key):
            super().setHandler(key, self.DEFAULT_HANDLER_CLASS())
        return super().getHandler(key)

    def setHandler(self, key, handler):
        return self.getHandler(key).use(handler)

    def clean(self, *args, **kwargs):
        self.cleaner.clean(*args, **kwargs)

    def onNew(self, sid):  # pylint: disable=unused-argument
        return self._createNew(self._genid())

    def onData(self, sid, id, data):
        self.clean()

        stream = self.streams.get(id)
        if not stream:
            stream = Stream(id).update(event=data["event"], **data["info"])
            self.streams[id] = stream
        stream.write(data["chunk"])

        payload = None
        if self.hasHandler(stream.event):
            payload = self._handle(stream.event, self.getHandler(stream.event))(
                sid, stream.data, stream
            )

        if stream.done:
            self._done(id, {"total": stream.total, "payload": payload})
        else:
            self.more(id, stream.total)

    def onStop(self, sid, id):  # pylint: disable=unused-argument
        if id in self.streams:
            self._done(id)

    def onResume(self, sid, id):  # pylint: disable=unused-argument
        stream = self.streams.get(id)
        if stream:
            self.resume(id, stream.total)
        else:
            self._createNew(id)
            self.resume(id)

    def more(self, id, *args, **kwargs):
        self.sio.emit(self.event.more(id), *args, **kwargs)

    def resume(self, id, *args, **kwargs):
        self.sio.emit(self.event.resume(id), *args, **kwargs)

    def end(self, id, *args, **kwargs):
        self.sio.emit(self.event.end(id), *args, **kwargs)


class WebSocketHandler(geventwebsocket.handler.WebSocketHandler):
    def get_environ(self):
        env = super().get_environ()
        urlpath = self.path.split("?", 1)[0] if "?" in self.path else self.path
        env["PATH_INFO"] = urllib.parse.unquote(urlpath)
        return env



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import g
from suanpan.app import app
from suanpan.app.modules.base import Module

module = Module()


@module.fn
def notifyError(err, *args, **kwargs):
    app.sio.error(err, *args, **kwargs)


@module.on("core.globals")
def getImageMeta(context):
    args = context.args
    return g.json(args["keys"])



# coding=utf-8
from __future__ import absolute_import, print_function

import suanpan
from suanpan.app import app
from suanpan.app.arguments import String
from suanpan.dw import dw
from suanpan.utils import csv


@app.param(String(key="file", required=True))
@app.param(String(key="table", required=True))
def SPDWDownload(context):
    args = context.args
    csv.dump(dw.readTable(args.table), args.file)


if __name__ == "__main__":
    suanpan.run(app)



# coding=utf-8
from __future__ import absolute_import, print_function

import base64
import hashlib
import hmac

from suanpan import error, g, utils


def signatureV1(secret, data):
    h = hmac.new(utils.encode(secret), utils.encode(data), hashlib.sha1)
    return utils.decode(base64.b64encode(utils.encode(h.digest())))


def defaultHeaders():
    if not g.userIdHeaderField:
        raise error.ApiError("UserIdHeaderField not set")
    return {
        g.userIdHeaderField: g.userId,
        g.userSignatureHeaderField: signatureV1(g.accessSecret, g.userId),
        g.userSignVersionHeaderField: "v1",
    }



# coding=utf-8
from __future__ import absolute_import, print_function

from addict import Dict


class HasName(object):
    @property
    def name(self):
        return self.__class__.__name__


class Context(Dict, HasName):
    @classmethod
    def froms(cls, *args, **kwargs):
        context = cls()
        context.update(item for arg in args for item in arg.items())
        context.update(kwargs)
        return context

    def __getitem__(self, name):
        value = super().__getitem__(name)
        if isinstance(value, property):
            value = value.fget(self)
            self.__setitem__(name, value)
        return value

    def json(self, keys=None):
        keys = keys or self.keys()
        return {key: self[key] for key in keys if self._isJsonable(key)}

    def _isValid(self, key):
        try:
            self.__getitem__(key)
            return True
        except Exception:  # pylint: disable=broad-except
            return False

    def _isJsonable(self, key):
        jsonableTypes = (str, int, float, bool, dict, list)
        return self._isValid(key) and isinstance(self[key], jsonableTypes)



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import path, runtime
from suanpan.state.base import IndexSaver, TimeSaver
from suanpan.state.storage import StorageSaver
from suanpan.storage import storage
from suanpan.utils import image


class ScreenshotsSaver(StorageSaver):
    def update(self, imageOrPath):
        data = image.read(imageOrPath) if isinstance(imageOrPath, str) else imageOrPath
        self.current.localPath = storage.getPathInTempStore(self.current.storageName)
        storageName = storage.storagePathJoin(
            self.current.storageName, self.currentPattern
        )
        localPath = storage.localPathJoin(self.current.localPath, self.currentPattern)
        image.save(localPath, data, self.current.get("flag"))
        return storage.upload(storageName, localPath)


class ScreenshotsWithThumbnailSaver(StorageSaver):
    WIDTH = 74

    def update(self, imageOrPath):
        data = image.read(imageOrPath) if isinstance(imageOrPath, str) else imageOrPath
        thumbnailData = self.resize(data)

        self.saveScreenshot(
            thumbnailData,
            self.current.thumbnail,
            self.currentPattern,
            self.current.get("flag"),
        )

        return self.saveScreenshot(
            data,
            self.current.storageName,
            self.currentPattern,
            self.current.get("flag"),
        )

    def clean(self):
        for storageName in [self.current.storageName, self.current.thumbnail]:
            self._clean(storageName)

    def resize(self, data):
        height, width = data.shape[0], data.shape[1]
        ratio = float(width) / self.WIDTH
        if ratio <= 1:
            return data

        return image.resize(data, size=(int(width / ratio), int(height / ratio)))

    def saveScreenshot(self, data, storageName, pattern, flag):
        tempPath = storage.getPathInTempStore(storageName)
        storageName = storage.storagePathJoin(storageName, pattern)
        localPath = storage.localPathJoin(tempPath, pattern)

        image.save(localPath, data, flag)
        storage.upload(storageName, localPath)

    def _clean(self, storageName):
        localPath = storage.getPathInTempStore(storageName)
        runtime.saferun(storage.remove)(storageName)
        return path.remove(localPath)


class ScreenshotsIndexSaver(ScreenshotsWithThumbnailSaver, IndexSaver):
    PATTERN = "screenshot_{index}.png"


class ScreenshotsTimeSaver(ScreenshotsWithThumbnailSaver, TimeSaver):
    PATTERN = "screenshot_{time}.png"



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import error, g
from suanpan.api import requests


def getHostUrl(path, tls=g.hostTls):
    if not g.host:
        raise error.ApiError("host not set")
    protocol = "https" if tls else "http"
    return f"{protocol}://{g.host}{path}"


def getApiHostUrl(appId, apiName, tls=g.hostApiTls):
    if not g.apiHost:
        raise error.ApiError("ApiHost not set")
    protocol = "https" if tls else "http"
    return f"{protocol}://{g.apiHost}/{appId}/{apiName}"


def get(path, *args, **kwargs):
    return requests.get(getHostUrl(path), *args, **kwargs)


def post(path, *args, **kwargs):
    return requests.post(getHostUrl(path), *args, **kwargs)


def put(path, *args, **kwargs):
    return requests.put(getHostUrl(path), *args, **kwargs)


def delete(path, *args, **kwargs):
    return requests.delete(getHostUrl(path), *args, **kwargs)


def call(appId, apiName, data, *args, **kwargs):
    kwargs.setdefault("auth", False)
    debug = kwargs.pop("debug", False)
    return requests.post(
        getApiHostUrl(appId, apiName),
        json={"data": data, "debug": debug},
        *args,
        **kwargs,
    )
